;** Part of this code is taken from libav: http://www.libav.org ***************
;** original file:
;**   libswscale/x86/input.asm
;** UV_TO_NVXX_FN by xv
;******************************************************************************
;* x86-optimized input routines; does shuffling of packed
;* YUV formats into individual planes, and converts RGB
;* into YUV planes also.
;* Copyright (c) 2012 Ronald S. Bultje <rsbultje@gmail.com>
;*
;* This file is part of Libav.
;*
;* Libav is free software; you can redistribute it and/or
;* modify it under the terms of the GNU Lesser General Public
;* License as published by the Free Software Foundation; either
;* version 2.1 of the License, or (at your option) any later version.
;*
;* Libav is distributed in the hope that it will be useful,
;* but WITHOUT ANY WARRANTY; without even the implied warranty of
;* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;* Lesser General Public License for more details.
;*
;* You should have received a copy of the GNU Lesser General Public
;* License along with Libav; if not, write to the Free Software
;* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
;******************************************************************************

%define ARCH_X86 1
%define ARCH_X86_32 0
%define ARCH_X86_64 1

%include "x86inc.asm"

SECTION_RODATA

umask16: ddq 0x80078006800580048003800280018000
vmask16: ddq 0x07800680058004800380028001800080

umask8: dq 0x8003800280018000
vmask8: dq 0x0380028001800080

SECTION .text

; %1 = a (aligned) or u (unaligned)
; %2 = nv12 or nv21
%macro LOOP_NVXX_TO_UV 2
.loop_%1:
    mov%1          m0, [srcq+wq*2]        ; (byte) { U0, V0, U1, V1, ... }
    mov%1          m1, [srcq+wq*2+mmsize] ; (byte) { U8, V8, U9, V9, ... }
    pand           m2, m0, m4             ; (word) { U0, U1, ..., U7 }
    pand           m3, m1, m4             ; (word) { U8, U9, ..., U15 }
    psrlw          m0, 8                  ; (word) { V0, V1, ..., V7 }
    psrlw          m1, 8                  ; (word) { V8, V9, ..., V15 }
    packuswb       m2, m3                 ; (byte) { U0, ..., U15 }
    packuswb       m0, m1                 ; (byte) { V0, ..., V15 }
%ifidn %2, nv12
    mova   [dstUq+wq], m2
    mova   [dstVq+wq], m0
%else ; nv21
    mova   [dstVq+wq], m2
    mova   [dstUq+wq], m0
%endif ; nv12/21
    add            wq, mmsize
    jl .loop_%1
    REP_RET
%endmacro

; %1 = nr. of XMM registers
; %2 = nv12 or nv21
%macro NVXX_TO_UV_FN 2
cglobal %2ToUV, 3, 4, %1, dstU, dstV, src, w
%if ARCH_X86_64
    movsxd         wq, dword r3m
%else ; x86-32
    mov            wq, r3m
%endif
    add         dstUq, wq
    add         dstVq, wq
%if mmsize == 16
    test         srcq, 15
%endif
    lea          srcq, [srcq+wq*2]
    pcmpeqb        m4, m4                 ; (byte) { 0xff } x 16
    psrlw          m4, 8                  ; (word) { 0x00ff } x 8
%if mmsize == 16
    jnz .loop_u_start
    neg            wq
    LOOP_NVXX_TO_UV a, %2
.loop_u_start:
    neg            wq
    LOOP_NVXX_TO_UV u, %2
%else ; mmsize == 8
    neg            wq
    LOOP_NVXX_TO_UV a, %2
%endif ; mmsize == 8/16
%endmacro


%macro LOOP_UV_TO_NVXX 2
.loop_%1:
%if mmsize == 16
    movq           m0, [srcUq+wq]         ; (byte) { U0, U1, ... }
    movq           m1, [srcVq+wq]         ; (byte) { V0, V1, ... }
%ifidn %2, nv12
    pshufb         m0, [umask16]          ; (byte) { U0, 00, U1, 00, ... }
    pshufb         m1, [vmask16]          ; (byte) { 00, V0, 00, V1, ... }
%else ; nv21
    pshufb         m0, [vmask16]
    pshufb         m1, [umask16]
%endif ; nv12/21
%else
    movd           m0, [srcUq+wq]         ; (byte) { U0, U1, ... }
    movd           m1, [srcVq+wq]         ; (byte) { V0, V1, ... }
%ifidn %2, nv12
    pshufb         m0, [umask8]           ; (byte) { U0, 00, U1, 00, ... }
    pshufb         m1, [vmask8]           ; (byte) { 00, V0, 00, V1, ... }
%else ; nv21
    pshufb         m0, [vmask8]
    pshufb         m1, [umask8]
%endif ; nv12/21
%endif
    por            m0, m1                 ; (byte) { U0, V0, U1, V1, ... }
    mov%1 [dstq+wq*2], m0
%if mmsize == 16
    add            wq, 8
%else
    add            wq, 4
%endif
    jl .loop_%1
    REP_RET
%endmacro

; %1 = nr. of XMM registers
; %2 = nv12 or nv21
%macro UV_TO_NVXX_FN 2
cglobal UVTo%2, 3, 4, %1, dst, srcU, srcV, w
%if ARCH_X86_64
    movsxd         wq, dword r3m
%else ; x86-32
    mov            wq, r3m
%endif
    add         srcUq, wq
    add         srcVq, wq
    lea          dstq, [dstq+wq*2]
%if mmsize == 16
    test         dstq, 15
    jnz .loop_u_start
    neg            wq
    LOOP_UV_TO_NVXX a, %2
.loop_u_start:
    neg            wq
    LOOP_UV_TO_NVXX u, %2
%else ; mmsize == 8
    neg            wq
    LOOP_UV_TO_NVXX a, %2
%endif ; mmsize == 8/16
%endmacro

INIT_XMM sse2
NVXX_TO_UV_FN 5, nv12

INIT_XMM ssse3
UV_TO_NVXX_FN 2, nv12