/*
 * Copyright (C) 2006-2010 Michael Niedermayer <michaelni@gmx.at>
 *               2010      James Darnley <james.darnley@gmail.com>
 *
 * This file is part of Libav.
 *
 * Libav is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Libav; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

//#include "libavutil/cpu.h"
//#include "libavutil/common.h"
//#include "libavutil/pixdesc.h"
//#include "avfilter.h"
//#include "video.h"
#include <libintl.h>
#define _(str) gettext(str)		///< gettext shortcut
#define _N(str) str			///< gettext_noop shortcut
#include "vaapi_yadif.h"
#include "yadif.h"
#include "misc.h"
/*
#undef NDEBUG
#include <assert.h>

typedef struct {
    / **
     * 0: send 1 frame for each frame
     * 1: send 1 frame for each field
     * 2: like 0 but skips spatial interlacing check
     * 3: like 1 but skips spatial interlacing check
     * /
    int mode;

    / **
     *  0: top field first
     *  1: bottom field first
     * -1: auto-detection
     * /
    int parity;

    int frame_pending;

    / **
     *  0: deinterlace all frames
     *  1: only deinterlace frames marked as interlaced
     * /
    int auto_enable;

    AVFilterBufferRef *cur;
    AVFilterBufferRef *next;
    AVFilterBufferRef *prev;
    AVFilterBufferRef *out;
    void (*filter_line)(uint8_t *dst,
                        uint8_t *prev, uint8_t *cur, uint8_t *next,
                        int w, int prefs, int mrefs, int parity, int mode);

    const AVPixFmtDescriptor *csp;
    int eof;
} YADIFContext;*/

#define CHECK(j)\
    {   int score = FFABS(cur[mrefs-1+(j)] - cur[prefs-1-(j)])\
                  + FFABS(cur[mrefs  +(j)] - cur[prefs  -(j)])\
                  + FFABS(cur[mrefs+1+(j)] - cur[prefs+1-(j)]);\
        if (score < spatial_score) {\
            spatial_score= score;\
            spatial_pred= (cur[mrefs  +(j)] + cur[prefs  -(j)])>>1;\

#define FILTER \
    for (x = 0;  x < w; x++) { \
        int c = cur[mrefs]; \
        int d = (prev2[0] + next2[0])>>1; \
        int e = cur[prefs]; \
        int temporal_diff0 = FFABS(prev2[0] - next2[0]); \
        int temporal_diff1 =(FFABS(prev[mrefs] - c) + FFABS(prev[prefs] - e) )>>1; \
        int temporal_diff2 =(FFABS(next[mrefs] - c) + FFABS(next[prefs] - e) )>>1; \
        int diff = FFMAX3(temporal_diff0>>1, temporal_diff1, temporal_diff2); \
        int spatial_pred = (c+e)>>1; \
        int spatial_score = FFABS(cur[mrefs-1] - cur[prefs-1]) + FFABS(c-e) \
                          + FFABS(cur[mrefs+1] - cur[prefs+1]) - 1; \
 \
        CHECK(-1) CHECK(-2) }} }} \
        CHECK( 1) CHECK( 2) }} }} \
 \
        if (mode < 2) { \
            int b = (prev2[2*mrefs] + next2[2*mrefs])>>1; \
            int f = (prev2[2*prefs] + next2[2*prefs])>>1; \
            int max = FFMAX3(d-e, d-c, FFMIN(b-c, f-e)); \
            int min = FFMIN3(d-e, d-c, FFMAX(b-c, f-e)); \
 \
            diff = FFMAX3(diff, min, -max); \
        } \
 \
        if (spatial_pred > d + diff) \
           spatial_pred = d + diff; \
        else if (spatial_pred < d - diff) \
           spatial_pred = d - diff; \
 \
        dst[0] = spatial_pred; \
 \
        dst++; \
        cur++; \
        prev++; \
        next++; \
        prev2++; \
        next2++; \
    }

static void filter_line_c(uint8_t *dst,
                          uint8_t *prev, uint8_t *cur, uint8_t *next,
                          int w, int prefs, int mrefs, int parity, int mode)
{
    int x;
    uint8_t *prev2 = parity ? prev : cur ;
    uint8_t *next2 = parity ? cur  : next;

    FILTER
}

/*static void filter_line_c_16bit(uint16_t *dst,
                                uint16_t *prev, uint16_t *cur, uint16_t *next,
                                int w, int prefs, int mrefs, int parity, int mode)
{
    int x;
    uint16_t *prev2 = parity ? prev : cur ;
    uint16_t *next2 = parity ? cur  : next;
    mrefs /= 2;
    prefs /= 2;

    FILTER
}*/

static void nv12ToUV(uint8_t *dstU, uint8_t *dstV, uint8_t *src, int width)
{
	int i;
	//width>>=1;
	for(i=0;i<width;i++) {
		dstU[i]=src[i*2];
		dstV[i]=src[i*2+1];
	}
}

static void UVTonv12(uint8_t *dst, uint8_t *srcU, uint8_t *srcV, int width)
{
	int i;
	//width>>=1;
	for(i=0;i<width;i++) {
		dst[i*2]=srcU[i];
		dst[i*2+1]=srcV[i];
	}
}

#define YADIF_PITCH(w,a) (w%a==0) ? w : w+a-(w%a)

static void CopyVAImg2YadifImg_alloc(VADisplay *display, VAImage *src, vaapi_yadif_img *dst)
{
	uint8_t *srcp, *srcuv, *dstU, *dstV;
	int i;
	vaMapBuffer(display, src->buf, (void**)&srcp);
	dst->width[0] = src->width;
	dst->width[1] = dst->width[2] = dst->width[0]>>1;
	dst->height[0] = src->height;
	dst->height[1] = dst->height[2] = dst->height[0]>>1;
	if(src->format.fourcc == VA_FOURCC_NV12) {
		dst->pitch[0] = src->pitches[0];
		dst->pitch[2] = dst->pitch[1] = YADIF_PITCH(dst->width[1],16);
		dst->yuvsize = dst->pitch[0]*dst->height[0]+dst->pitch[1]*dst->height[1]+dst->pitch[2]*dst->height[2];
		dst->yuv[0] = malloc(dst->yuvsize);
		dst->yuv[1] = dst->yuv[0]+dst->pitch[0]*dst->height[0];
		dst->yuv[2] = dst->yuv[1]+dst->pitch[1]*dst->height[1];
		memcpy(dst->yuv[0], srcp+src->offsets[0], src->height*src->pitches[0]);
		srcuv=srcp+src->offsets[1];
		dstU=dst->yuv[1];
		dstV=dst->yuv[2];
		for(i=0;i<dst->height[1];i++) {
#ifdef HAVE_SSE2
			vdr_plg_shddev_nv12ToUV_sse2(dstU,dstV,srcuv,dst->width[1]);
#else
			nv12ToUV(dstU,dstV,srcuv,dst->width[1]);
#endif
			srcuv+=src->pitches[1];
			dstU+=dst->pitch[1];
			dstV+=dst->pitch[2];
		}
	}
	if (vaUnmapBuffer(display, src->buf) != VA_STATUS_SUCCESS) {
		Error(_("video/vaapi: can't unmap image buffer\n"));
    }
}

static void CopyVAImg2YadifImg(VADisplay *display, VAImage *src, vaapi_yadif_img *dst)
{
	uint8_t *srcp, *srcuv, *dstU, *dstV;
	int i;
	vaMapBuffer(display, src->buf, (void**)&srcp);
	if(src->format.fourcc == VA_FOURCC_NV12) {
		memcpy(dst->yuv[0], srcp+src->offsets[0], src->height*src->pitches[0]);
		srcuv=srcp+src->offsets[1];
		dstU=dst->yuv[1];
		dstV=dst->yuv[2];
		for(i=0;i<dst->height[1];i++) {
#ifdef HAVE_SSE2
			vdr_plg_shddev_nv12ToUV_sse2(dstU,dstV,srcuv,dst->width[1]);
#else
			nv12ToUV(dstU,dstV,srcuv,dst->width[1]);
#endif
			srcuv+=src->pitches[1];
			dstU+=dst->pitch[1];
			dstV+=dst->pitch[2];
		}
	}
	if (vaUnmapBuffer(display, src->buf) != VA_STATUS_SUCCESS) {
		Error(_("video/vaapi: can't unmap image buffer\n"));
    }
}

static void CopyYadifImg2VAImg(VADisplay *display, vaapi_yadif_img *src, VAImage *dst)
{
	uint8_t *dstp, *dstuv, *srcU, *srcV;
	int i;
	vaMapBuffer(display, dst->buf, (void**)&dstp);
	if(dst->format.fourcc == VA_FOURCC_NV12) {
		memcpy(dstp+dst->offsets[0], src->yuv[0], src->height[0]*src->pitch[0]);
		dstuv=dstp+dst->offsets[1];
		srcU=src->yuv[1];
		srcV=src->yuv[2];
		for(i=0;i<src->height[1];i++) {
#ifdef HAVE_SSSE3
			vdr_plg_shddev_UVTonv12_ssse3(dstuv,srcU,srcV,src->width[1]);
#else
			UVTonv12(dstuv,srcU,srcV,src->width[1]);
#endif
			dstuv+=dst->pitches[1];
			srcU+=src->pitch[1];
			srcV+=src->pitch[2];
		}
	}
	if (vaUnmapBuffer(display, dst->buf) != VA_STATUS_SUCCESS) {
		Error(_("video/vaapi: can't unmap image buffer\n"));
    }
}

static void InitVaapiYadif(VADisplay *display, VAImage *src, vaapi_yadif_img *buf_imgs[])
{
	buf_imgs[0]=malloc(5*sizeof(vaapi_yadif_img));
	//printf("init yadif1\n");
	buf_imgs[1]=((void*)buf_imgs[0])+sizeof(vaapi_yadif_img);
	buf_imgs[2]=((void*)buf_imgs[1])+sizeof(vaapi_yadif_img);
	buf_imgs[3]=((void*)buf_imgs[2])+sizeof(vaapi_yadif_img);
	buf_imgs[4]=((void*)buf_imgs[3])+sizeof(vaapi_yadif_img);
	//printf("init yadif2\n");
	CopyVAImg2YadifImg_alloc(display,src,buf_imgs[0]);
	//printf("init yadif3\n");
	memcpy(buf_imgs[1],buf_imgs[0],sizeof(vaapi_yadif_img));
	memcpy(buf_imgs[2],buf_imgs[0],sizeof(vaapi_yadif_img));
	buf_imgs[1]->yuv[0] = malloc(buf_imgs[1]->yuvsize);
	buf_imgs[2]->yuv[0] = malloc(buf_imgs[2]->yuvsize);
	memcpy(buf_imgs[1]->yuv[0],buf_imgs[0]->yuv[0],buf_imgs[1]->yuvsize);
	memcpy(buf_imgs[2]->yuv[0],buf_imgs[0]->yuv[0],buf_imgs[2]->yuvsize);
	buf_imgs[1]->yuv[1] = buf_imgs[1]->yuv[0]+buf_imgs[1]->pitch[0]*buf_imgs[1]->height[0];
	buf_imgs[1]->yuv[2] = buf_imgs[1]->yuv[1]+buf_imgs[1]->pitch[1]*buf_imgs[1]->height[1];
	buf_imgs[2]->yuv[1] = buf_imgs[2]->yuv[0]+buf_imgs[2]->pitch[0]*buf_imgs[2]->height[0];
	buf_imgs[2]->yuv[2] = buf_imgs[2]->yuv[1]+buf_imgs[2]->pitch[1]*buf_imgs[2]->height[1];
	memcpy(buf_imgs[3],buf_imgs[0],sizeof(vaapi_yadif_img));
	memcpy(buf_imgs[4],buf_imgs[0],sizeof(vaapi_yadif_img));
	buf_imgs[3]->yuv[0] = malloc(buf_imgs[3]->yuvsize);
	buf_imgs[4]->yuv[0] = malloc(buf_imgs[4]->yuvsize);
	buf_imgs[3]->yuv[1] = buf_imgs[3]->yuv[0]+buf_imgs[3]->pitch[0]*buf_imgs[3]->height[0];
	buf_imgs[3]->yuv[2] = buf_imgs[3]->yuv[1]+buf_imgs[3]->pitch[1]*buf_imgs[3]->height[1];
	buf_imgs[4]->yuv[1] = buf_imgs[4]->yuv[0]+buf_imgs[4]->pitch[0]*buf_imgs[4]->height[0];
	buf_imgs[4]->yuv[2] = buf_imgs[4]->yuv[1]+buf_imgs[4]->pitch[1]*buf_imgs[4]->height[1];
}

void VaapiYadif (VADisplay *display, VAImage *src, VAImage *dst1,
    VAImage *dst2, vaapi_yadif_img *buf_imgs[], int tff, int skip_spatial)
{
	/*buf_imgs[0]
	buf_imgs[1]*/
	//int y, i;
	vaapi_yadif_img *tmp;
	//int tff;
//	printf("ok 1\n");
if(buf_imgs[2]==NULL) { InitVaapiYadif(display, src, buf_imgs); printf("init\n"); }
	else { //printf("noinitheight: %i %i %i %i %i\n",buf_imgs[0]->height[0],buf_imgs[1]->height[0],buf_imgs[2]->height[0],buf_imgs[3]->height[0],buf_imgs[4]->height[0]);
//	printf("adress: %x %x %x %x %x\n",buf_imgs[0],buf_imgs[1],buf_imgs[2],buf_imgs[3],buf_imgs[4]);
//	printf("adresB: %x %x %x %x %x\n",buf_imgs[0]->height,buf_imgs[1]->height,buf_imgs[2]->height,buf_imgs[3]->height,buf_imgs[4]->height);
CopyVAImg2YadifImg(display, src, buf_imgs[2]); /* printf("noinit\n"); */ }
//	printf("ok 2\n");
//	printf("height: %i %i %i %i %i\n",buf_imgs[0]->height[0],buf_imgs[1]->height[0],buf_imgs[2]->height[0],buf_imgs[3]->height[0],buf_imgs[4]->height[0]);
	//tff = decoder->TopFieldFirst;
	
	skip_spatial<<=1;
	//parity = (mode & 1) ? (ndst & 1) ^ (1^tff) : (tff ^ 1);  // 0 or 1
	
	
	
	int y, i;

//	printf("ok 3\n");

    for (i = 0; i < 3; i++) {
        int w = buf_imgs[3]->width[i];
        int h = buf_imgs[3]->height[i];
        int refs = buf_imgs[3]->pitch[i];
        //int df = (yadif->csp->comp[i].depth_minus1 + 8) / 8;

        /*if (i == 1 || i == 2) {
          Why is this not part of the per-plane description thing? * /
            w >>= yadif->csp->log2_chroma_w;
            h >>= yadif->csp->log2_chroma_h;
        }*/
        //printf("height: %i %i %i %i %i\n",buf_imgs[0]->height[0],buf_imgs[1]->height[0],buf_imgs[2]->height[0],buf_imgs[3]->height[0],buf_imgs[4]->height[0]);
		//printf("ok p 1, plane %i height %i\n",i,h);
        for (y = 0; y < h; y++) {
        	//printf("ok L 1\n");
            if ((y ^ tff) & 1) {
            	//printf("ok L 1a\n");
                uint8_t *prev = &buf_imgs[0]->yuv[i][y*refs];
                uint8_t *cur  = &buf_imgs[1]->yuv[i][y*refs];
                uint8_t *next = &buf_imgs[2]->yuv[i][y*refs];
                //uint8_t *dst  = &buf_imgs[3]->yuv[i][y*refs];
                //int     mode  = y==1 || y+2==h ? 2 : yadif->mode;
                // vdr_plg_shddev_yadif_filter_line_ssse3
                filter_line_c(&buf_imgs[4]->yuv[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 0, (y==1 || y+2==h) ? 2 : skip_spatial);
                memcpy(&buf_imgs[3]->yuv[i][y*refs], &buf_imgs[1] ->yuv[i][y*refs], w);
                //printf("ok L 2a\n");
            } else {
            	//printf("ok L 1b\n");
                uint8_t *prev = &buf_imgs[0]->yuv[i][y*refs];
                uint8_t *cur  = &buf_imgs[1] ->yuv[i][y*refs];
                uint8_t *next = &buf_imgs[2]->yuv[i][y*refs];
                //uint8_t *dst  = &buf_imgs[3]->yuv[i][y*refs];
                //int     mode  = y==1 || y+2==h ? 2 : yadif->mode;
                filter_line_c(&buf_imgs[3]->yuv[i][y*refs], prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, 1, (y==1 || y+2==h) ? 2 : skip_spatial);
                memcpy(&buf_imgs[4]->yuv[i][y*refs], &buf_imgs[1] ->yuv[i][y*refs], w);
                //printf("ok L 2b\n");
            }
        }
    }
#if HAVE_MMX
    __asm__ volatile("emms \n\t" : : : "memory");
#endif
//	printf("height: %i %i %i %i %i\n",buf_imgs[0]->height[0],buf_imgs[1]->height[0],buf_imgs[2]->height[0],buf_imgs[3]->height[0],buf_imgs[4]->height[0]);
//	printf("ok 4\n");
	CopyYadifImg2VAImg(display,buf_imgs[3],dst1);
	CopyYadifImg2VAImg(display,buf_imgs[4],dst2);
//	printf("ok 5\n");
//	printf("height: %i %i %i %i %i\n",buf_imgs[0]->height[0],buf_imgs[1]->height[0],buf_imgs[2]->height[0],buf_imgs[3]->height[0],buf_imgs[4]->height[0]);
	tmp = buf_imgs[0];
	buf_imgs[0] = buf_imgs[1];
	buf_imgs[1] = buf_imgs[2];
	buf_imgs[2] = tmp;
//	printf("height: %i %i %i %i %i\n",buf_imgs[0]->height[0],buf_imgs[1]->height[0],buf_imgs[2]->height[0],buf_imgs[3]->height[0],buf_imgs[4]->height[0]);
//	printf("adress: %x %x %x %x %x\n",buf_imgs[0],buf_imgs[1],buf_imgs[2],buf_imgs[3],buf_imgs[4]);
//	printf("adresB: %x %x %x %x %x\n",buf_imgs[0]->height,buf_imgs[1]->height,buf_imgs[2]->height,buf_imgs[3]->height,buf_imgs[4]->height);
}

/*
static void filter(AVFilterContext *ctx, AVFilterBufferRef *dstpic,
                   int parity, int tff)
{
    YADIFContext *yadif = ctx->priv;
    int y, i;

    for (i = 0; i < yadif->csp->nb_components; i++) {
        int w = dstpic->video->w;
        int h = dstpic->video->h;
        int refs = yadif->cur->linesize[i];
        int df = (yadif->csp->comp[i].depth_minus1 + 8) / 8;

        if (i == 1 || i == 2) {
        /* Why is this not part of the per-plane description thing? * /
            w >>= yadif->csp->log2_chroma_w;
            h >>= yadif->csp->log2_chroma_h;
        }

        for (y = 0; y < h; y++) {
            if ((y ^ parity) & 1) {
                uint8_t *prev = &yadif->prev->data[i][y*refs];
                uint8_t *cur  = &yadif->cur ->data[i][y*refs];
                uint8_t *next = &yadif->next->data[i][y*refs];
                uint8_t *dst  = &dstpic->data[i][y*dstpic->linesize[i]];
                int     mode  = y==1 || y+2==h ? 2 : yadif->mode;
                yadif->filter_line(dst, prev, cur, next, w, y+1<h ? refs : -refs, y ? -refs : refs, parity ^ tff, mode);
            } else {
                memcpy(&dstpic->data[i][y*dstpic->linesize[i]],
                       &yadif->cur->data[i][y*refs], w*df);
            }
        }
    }
#if HAVE_MMX
    __asm__ volatile("emms \n\t" : : : "memory");
#endif
}


static AVFilterBufferRef *get_video_buffer(AVFilterLink *link, int perms, int w, int h)
{
    AVFilterBufferRef *picref;
    int width = FFALIGN(w, 32);
    int height= FFALIGN(h+2, 32);
    int i;

    picref = ff_default_get_video_buffer(link, perms, width, height);

    picref->video->w = w;
    picref->video->h = h;

    for (i = 0; i < 3; i++)
        picref->data[i] += picref->linesize[i];

    return picref;
}

static void return_frame(AVFilterContext *ctx, int is_second)
{
    YADIFContext *yadif = ctx->priv;
    AVFilterLink *link= ctx->outputs[0];
    int tff;

    if (yadif->parity == -1) {
        tff = yadif->cur->video->interlaced ?
            yadif->cur->video->top_field_first : 1;
    } else {
        tff = yadif->parity^1;
    }

    if (is_second) {
        yadif->out = avfilter_get_video_buffer(link, AV_PERM_WRITE | AV_PERM_PRESERVE |
                                               AV_PERM_REUSE, link->w, link->h);
        avfilter_copy_buffer_ref_props(yadif->out, yadif->cur);
        yadif->out->video->interlaced = 0;
    }

    if (!yadif->csp)
        yadif->csp = &av_pix_fmt_descriptors[link->format];
    if (yadif->csp->comp[0].depth_minus1 / 8 == 1)
        yadif->filter_line = filter_line_c_16bit;

    filter(ctx, yadif->out, tff ^ !is_second, tff);

    if (is_second) {
        int64_t cur_pts  = yadif->cur->pts;
        int64_t next_pts = yadif->next->pts;

        if (next_pts != AV_NOPTS_VALUE && cur_pts != AV_NOPTS_VALUE) {
            yadif->out->pts = cur_pts + next_pts;
        } else {
            yadif->out->pts = AV_NOPTS_VALUE;
        }
        avfilter_start_frame(ctx->outputs[0], yadif->out);
    }
    avfilter_draw_slice(ctx->outputs[0], 0, link->h, 1);
    avfilter_end_frame(ctx->outputs[0]);

    yadif->frame_pending = (yadif->mode&1) && !is_second;
}

static void start_frame(AVFilterLink *link, AVFilterBufferRef *picref)
{
    AVFilterContext *ctx = link->dst;
    YADIFContext *yadif = ctx->priv;

    if (yadif->frame_pending)
        return_frame(ctx, 1);

    if (yadif->prev)
        avfilter_unref_buffer(yadif->prev);
    yadif->prev = yadif->cur;
    yadif->cur  = yadif->next;
    yadif->next = picref;

    if (!yadif->cur)
        return;

    if (yadif->auto_enable && !yadif->cur->video->interlaced) {
        yadif->out  = avfilter_ref_buffer(yadif->cur, AV_PERM_READ);
        avfilter_unref_buffer(yadif->prev);
        yadif->prev = NULL;
        if (yadif->out->pts != AV_NOPTS_VALUE)
            yadif->out->pts *= 2;
        avfilter_start_frame(ctx->outputs[0], yadif->out);
        return;
    }

    if (!yadif->prev)
        yadif->prev = avfilter_ref_buffer(yadif->cur, AV_PERM_READ);

    yadif->out = avfilter_get_video_buffer(ctx->outputs[0], AV_PERM_WRITE | AV_PERM_PRESERVE |
                                       AV_PERM_REUSE, link->w, link->h);

    avfilter_copy_buffer_ref_props(yadif->out, yadif->cur);
    yadif->out->video->interlaced = 0;
    if (yadif->out->pts != AV_NOPTS_VALUE)
        yadif->out->pts *= 2;
    avfilter_start_frame(ctx->outputs[0], yadif->out);
}

static void end_frame(AVFilterLink *link)
{
    AVFilterContext *ctx = link->dst;
    YADIFContext *yadif = ctx->priv;

    if (!yadif->out)
        return;

    if (yadif->auto_enable && !yadif->cur->video->interlaced) {
        avfilter_draw_slice(ctx->outputs[0], 0, link->h, 1);
        avfilter_end_frame(ctx->outputs[0]);
        return;
    }

    return_frame(ctx, 0);
}

static int request_frame(AVFilterLink *link)
{
    AVFilterContext *ctx = link->src;
    YADIFContext *yadif = ctx->priv;

    if (yadif->frame_pending) {
        return_frame(ctx, 1);
        return 0;
    }

    do {
        int ret;

        if (yadif->eof)
            return AVERROR_EOF;

        ret  = avfilter_request_frame(link->src->inputs[0]);

        if (ret == AVERROR_EOF && yadif->next) {
            AVFilterBufferRef *next = avfilter_ref_buffer(yadif->next, AV_PERM_READ);
            next->pts = yadif->next->pts * 2 - yadif->cur->pts;

            start_frame(link->src->inputs[0], next);
            end_frame(link->src->inputs[0]);
            yadif->eof = 1;
        } else if (ret < 0) {
            return ret;
        }
    } while (!yadif->cur);

    return 0;
}

static int poll_frame(AVFilterLink *link)
{
    YADIFContext *yadif = link->src->priv;
    int ret, val;

    if (yadif->frame_pending)
        return 1;

    val = avfilter_poll_frame(link->src->inputs[0]);
    if (val <= 0)
        return val;

    if (val==1 && !yadif->next) { //FIXME change API to not requre this red tape
        if ((ret = avfilter_request_frame(link->src->inputs[0])) < 0)
            return ret;
        val = avfilter_poll_frame(link->src->inputs[0]);
        if (val <= 0)
            return val;
    }
    assert(yadif->next || !val);

    if (yadif->auto_enable && yadif->next && !yadif->next->video->interlaced)
        return val;

    return val * ((yadif->mode&1)+1);
}

static av_cold void uninit(AVFilterContext *ctx)
{
    YADIFContext *yadif = ctx->priv;

    if (yadif->prev) avfilter_unref_buffer(yadif->prev);
    if (yadif->cur ) avfilter_unref_buffer(yadif->cur );
    if (yadif->next) avfilter_unref_buffer(yadif->next);
}

static int query_formats(AVFilterContext *ctx)
{
    static const enum PixelFormat pix_fmts[] = {
        PIX_FMT_YUV420P,
        PIX_FMT_YUV422P,
        PIX_FMT_YUV444P,
        PIX_FMT_YUV410P,
        PIX_FMT_YUV411P,
        PIX_FMT_GRAY8,
        PIX_FMT_YUVJ420P,
        PIX_FMT_YUVJ422P,
        PIX_FMT_YUVJ444P,
        AV_NE( PIX_FMT_GRAY16BE, PIX_FMT_GRAY16LE ),
        PIX_FMT_YUV440P,
        PIX_FMT_YUVJ440P,
        AV_NE( PIX_FMT_YUV420P10BE, PIX_FMT_YUV420P10LE ),
        AV_NE( PIX_FMT_YUV422P10BE, PIX_FMT_YUV422P10LE ),
        AV_NE( PIX_FMT_YUV444P10BE, PIX_FMT_YUV444P10LE ),
        AV_NE( PIX_FMT_YUV420P16BE, PIX_FMT_YUV420P16LE ),
        AV_NE( PIX_FMT_YUV422P16BE, PIX_FMT_YUV422P16LE ),
        AV_NE( PIX_FMT_YUV444P16BE, PIX_FMT_YUV444P16LE ),
        PIX_FMT_YUVA420P,
        PIX_FMT_NONE
    };

    avfilter_set_common_formats(ctx, avfilter_make_format_list(pix_fmts));

    return 0;
}

static av_cold int init(AVFilterContext *ctx, const char *args, void *opaque)
{
    YADIFContext *yadif = ctx->priv;
    int cpu_flags = av_get_cpu_flags();

    yadif->mode = 0;
    yadif->parity = -1;
    yadif->auto_enable = 0;
    yadif->csp = NULL;

    if (args) sscanf(args, "%d:%d:%d", &yadif->mode, &yadif->parity, &yadif->auto_enable);

    yadif->filter_line = filter_line_c;
    if (HAVE_SSSE3 && cpu_flags & AV_CPU_FLAG_SSSE3)
        yadif->filter_line = ff_yadif_filter_line_ssse3;
    else if (HAVE_SSE && cpu_flags & AV_CPU_FLAG_SSE2)
        yadif->filter_line = ff_yadif_filter_line_sse2;
    else if (HAVE_MMX && cpu_flags & AV_CPU_FLAG_MMX)
        yadif->filter_line = ff_yadif_filter_line_mmx;

    av_log(ctx, AV_LOG_INFO, "mode:%d parity:%d auto_enable:%d\n", yadif->mode, yadif->parity, yadif->auto_enable);

    return 0;
}

static void null_draw_slice(AVFilterLink *link, int y, int h, int slice_dir) { }

static int config_props(AVFilterLink *link)
{
    link->time_base.num = link->src->inputs[0]->time_base.num;
    link->time_base.den = link->src->inputs[0]->time_base.den * 2;
    link->w             = link->src->inputs[0]->w;
    link->h             = link->src->inputs[0]->h;

    return 0;
}

AVFilter avfilter_vf_yadif = {
    .name          = "yadif",
    .description   = NULL_IF_CONFIG_SMALL("Deinterlace the input image"),

    .priv_size     = sizeof(YADIFContext),
    .init          = init,
    .uninit        = uninit,
    .query_formats = query_formats,

    .inputs    = (AVFilterPad[]) {{ .name             = "default",
                                    .type             = AVMEDIA_TYPE_VIDEO,
                                    .start_frame      = start_frame,
                                    .get_video_buffer = get_video_buffer,
                                    .draw_slice       = null_draw_slice,
                                    .end_frame        = end_frame, },
                                  { .name = NULL}},

    .outputs   = (AVFilterPad[]) {{ .name             = "default",
                                    .type             = AVMEDIA_TYPE_VIDEO,
                                    .poll_frame       = poll_frame,
                                    .request_frame    = request_frame,
                                    .config_props     = config_props, },
                                  { .name = NULL}},
};
*/